(** General pretty printing functions *)

open Format;;

open Trace;;
open Abstract;;
open Symbols;;
open Pp;;
open Debugpp;;

open Clause;;


(***************************************************************
 * Separators for pp_list
 ***************************************************************)

(* standard comma separation to use for pp_list *)
let comma_sep () = 
  print_string ","; print_space();;

(* standard semicolumn separation to use for pp_list *)
let semicol_sep () = 
  print_string ";"; print_space();;

(* standard null separation to use for pp_list *)
let null_sep () = ();;

(* standard new line separation to use for pp_list *)
let newline_sep () = 
  print_newline ();;

(* double new line separation to use for pp_list *)
let newline2_sep () = 
  print_newline (); print_newline ();;


(***************************************************************
 * Pretty printing functions for TPTP format
 ***************************************************************)

let rec hornpp_id sl = function
    EVA_ID s -> hornpp_var sl s
      (* print_string ("X_"^s) *)
and hornpp_lid sl { lid_id = i }  =
  hornpp_id sl i
and hornpp_var sl s = 
  try
    let sy = assoc_symbol s sl in 
      if ((is_first_order sy) || (is_functional sy) || (is_key sy))
      then match (get_scope sy) with
          EVA_CST -> print_string ("c_"^s)
        | EVA_PARAM -> print_string ("P_"^s)
        | EVA_PRIVATE -> print_string ("V_"^s)
      else raise (Internal_error ("hornpp_var: "^s))
  with
      Symbols_not_found ->  
        raise (Internal_error 
                 ("hornpp_var: "^s^" Symbol_not_found"))
    | Symbols_error msg -> 
        raise (Internal_error 
                 ("hornpp_var: "^s^" Symbols_error "^msg));;

let rec hornpp_label = function
    EVA_NOLABEL 
  | EVA_LABEL "" -> 
      print_string "*nolab*"
  | EVA_LABEL s -> 
      (* remove the final '.' *)
      let last = (String.length s) -1 in
        if ((String.get s last) = '.')
        then print_string (String.sub s 0 last)
        else print_string s
and hornpp_llabel { lla_label = l } =
  hornpp_label l;;

let rec hornpp_type t =
  hornpp_gtype "t" t
and hornpp_ltype { lty_loc = _; lty_type = t } =
  hornpp_type t
(* prints type or generic type *)
and hornpp_gtype prefix t =
  open_box 0;
  (match t with
       EVA_VOID -> print_string (prefix^"_VOID")
     | EVA_PRINCIPAL -> print_string (prefix^"_principal")
     | EVA_NUMBER -> print_string (prefix^"_number")
     | EVA_AALGO -> print_string (prefix^"_asymalgo")
     | EVA_SALGO -> print_string (prefix^"_symalgo")
     | EVA_TALGO -> print_string (prefix^"_algo")
     | (EVA_USERTYPE s) -> print_string (prefix^"_"^s)
     | (EVA_BASETYPE s) -> print_string (prefix^"_"^s)
     | (EVA_FUN t1) -> hornpp_gtype prefix t1; print_string ("_number")
     | (EVA_HFUN t1) -> hornpp_gtype prefix t1; print_string ("_hnumber")
     | (EVA_KEY t1) -> hornpp_gtype prefix t1; print_string ("_kp")
  );
  close_box ();;

let rec hornpp_scope s =
  begin
    open_box 0;
    (match s with
       (*  EVA_SCOPE_VOID -> print_string "" *)
         EVA_CST -> print_string "constant"
       | EVA_PARAM -> print_string "parameter"
       | EVA_PRIVATE -> print_string "private");
    close_box ();
  end
and hornpp_lscope { lsc_loc = _; lsc_scope = s } =
  hornpp_scope s;;

let rec hornpp_value = function
    EVA_INTRUDER -> print_string "I"
  | EVA_VALUE s -> print_string ("v_"^s);
and hornpp_lvalue { lva_value = v } =
   hornpp_value v;;

let rec hornpp_term sl term =
  begin
    open_box 2;
    (match term with
         (* EVA_TERM_VOID -> print_string "" *)
         EVA_TERM_ID s -> 
           hornpp_var sl s
       | EVA_TERM_CST s -> 
           hornpp_var sl s
       | (EVA_TERM_VALUE v) -> 
           hornpp_value v
       | (EVA_TERM_ALIAS (a, ta)) -> 
           hornpp_lterm sl ta
       | (EVA_LOCATED (i, r)) -> 
           (* the variable is a parameter *)
           hornpp_lid sl i; 
       | (EVA_QUANTIFIED (i, _)) -> 
	   (* i is not in sl *)
           hornpp_termquantified i; 
       | EVA_P t ->
           print_string "r_p(";
           hornpp_lterm sl t;
           print_string ")";           
       | EVA_A t ->
           print_string "r_a(";
           hornpp_lterm sl t;
           print_string ")";           
       | EVA_SA t ->
           print_string "r_sa(";
           hornpp_lterm sl t;
           print_string ")";           
       | EVA_AA t ->
           print_string "r_aa(";
           hornpp_lterm sl t;
           print_string ")";           
       | EVA_U (i, t) ->
           print_string ("r_u_"^i^"(");
           hornpp_lterm sl t;
           print_string ")";           
       | EVA_CONS tl -> 
(*         pp_list (hornpp_lterm sl) comma_sep tl  *)
           hornpp_list sl tl
       | EVA_APP (i, etl) -> 
           hornpp_app sl (i, etl)
       | EVA_CRYPT (a, k, t) 
       | EVA_SIGN (a, k, t) -> 
           hornpp_cipher sl (a, k, t) 
       | EVA_PCENT (t1, t2) -> 
           raise (Internal_error "hornpp_term: %")
    );
    close_box ();
  end
and hornpp_lterm sl { lte_term = t } =
  hornpp_term sl t
(* app_root = i; app_args = etl *)
and hornpp_termid s =
  print_string ("X_"^s)
and hornpp_termcst s =
  print_string ("c_"^s)
and hornpp_termquantified s =
  print_string ("Q_"^s)
and hornpp_list sl = function
    [] -> ()
  | [t] -> hornpp_lterm sl t
  | l -> hornpp_listt sl l
and hornpp_listt sl = function
    [] -> 
      hornpp_value empty_list
  | t::l -> 
      print_string "c(";
      hornpp_lterm sl t;
      print_string ","; print_space ();
      hornpp_listt sl l;
      print_string ")";
and hornpp_app sl (i, etl) =
  print_string ("app"^(string_of_int (List.length etl)));
  print_string "(";
  hornpp_lid sl i; 
  print_string ","; print_space ();
  (* pp_list (hornpp_lterm sl) comma_sep etl;  *)
  hornpp_list sl etl;
  print_string ")";                 
(* crypt_algo = a; crypt_key = k; crypt_content = t *)
and hornpp_cipher sl (a, k, t) =
  print_string "crypt(";
  hornpp_algo sl a;
  print_string ","; print_space ();
  hornpp_lterm sl k;
  print_string ","; print_space ();
  hornpp_lterm sl t;
  print_string ")";                 
and hornpp_algo sl = function
    EVA_VANILLA -> print_string "vanilla"
  | (EVA_ALGO t) -> hornpp_lterm sl t;;

let rec hornpp_literal sign sl = function  
    STATE { st_role = r; st_label = l; st_args = al } -> 
      open_box 0;
      print_string sign;
      print_string "state_";
      print_string r;
      print_string "_";
      hornpp_label (EVA_LABEL l); (* dirty *)
      print_string "(";
      pp_list (hornpp_term sl) comma_sep al;
      print_string ")";
      close_box ()
  | INTRUK t ->
      open_box 0;
      print_string sign;
      print_string "knows(";
      hornpp_term sl t;
      print_string ")";
      close_box ()
  | HONEST t -> 
      open_box 0;
      print_string sign;
      print_string "honest(";
      hornpp_term sl t;
      print_string ")";
      close_box ()
  | TYPE (t, arg) -> 
      open_box 0;
      print_string sign;
      hornpp_type t;
      print_string "(";
      hornpp_term sl arg;
      print_string ")";
      close_box ()
  (* we treat algos like terms *)
  | TYPEA (t, arg) -> 
      open_box 0;
      print_string sign;
      hornpp_type t;
      print_string "(";
      hornpp_algo sl arg;
      print_string ")";
      close_box ()
  | TYPEG (t, arg) -> 
      open_box 0;
      print_string sign;
      hornpp_gtype "dt" t;
      print_string "(";
      hornpp_term sl arg;
      print_string ")";
      close_box ()
  | USER (i, t) -> 
      open_box 0;
      print_string sign;
      print_string (i^"(");
      hornpp_term sl t;
      print_string ")";
      close_box ();;


(***************************************************************
 * Pretty printing functions for folkloric clause format
 ***************************************************************)

let rec hornpp_clause sl { cl_neg = negl; cl_pos = posl } =
  open_box 2;
  pp_list (hornpp_literal "" sl) comma_sep negl;
  print_space (); print_string "=>"; print_space ();
  pp_list (hornpp_literal "" sl) comma_sep posl;
  close_box ()
and hornpp_tclause sl = function
    AXIOM c 
  | CONJ c ->
      hornpp_clause sl c
  | COMMENT s ->
      print_newline ();      
      print_string ("// "^s);
      print_newline ();; 

      
(* 
 * let hornpp_spec sp = 
 *   let sl = symbollist_of_spec sp in
 *   let clauses = horn_spec sl sp in
 *     pp_list (hornpp_tclause sl) newline2_sep clauses;
 *    print_newline ();; *)

